/* 
   - The contents of this file are subject to the Mozilla Public
   - License Version 1.1 (the "License"); you may not use this file
   - except in compliance with the License. You may obtain a copy of
   - the License at http://www.mozilla.org/MPL/
   -
   - Software distributed under the License is distributed on an "AS
   - IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
   - implied. See the License for the specific language governing
   - rights and limitations under the License.
   -
   - The Original Code is this file as it was released on
   - March 20, 2001.
   -
   - The Initial Developer of the Original Code is HJ van Rantwijk.
   - Portions created by HJ van Rantwijk are Copyright (C) 2001-2007
   - HJ van Rantwijk.  All Rights Reserved.
   -
   - Contributor(s):
   -  David Perry <d.perry@utoronto.ca> (limited to function preventDupeTab)
   -  Michael Vincent van Rantwijk <mv_van_rantwijk@yahoo.com>
   -
*/


function hideTab(aTab)
{
  var clsTab = "close";
  if (aTab.hasAttribute(clsTab))
    aTab.removeAttribute(clsTab);
  else aTab.setAttribute(clsTab, "true");
}

function loadFirstTabPage(aTab)
{
  if (aTab == null)
    aTab = gBrowser.selectedTab;

  var url = aTab.getAttribute("firstPage");
  var browser = aTab.linkedBrowser;

  if (browser && url != browser.currentURI.spec)
    browser.loadURI(url);
}

function FillSelectedHistoryMenu(aParent, aMenu, aTab)
{
  var sessionHistory = getWebNavigation().sessionHistory;
  var end;
  var j;
  var entry;

  if (aTab) {
    var browser = aTab.linkedBrowser;
    sessionHistory = browser.sessionHistory;
  }
  var count = sessionHistory.count;
  var index = sessionHistory.index;

  deleteHistoryItems(aParent);

  if (aMenu == "global" && count == 0)
    return;

  switch (aMenu)
    {
      case "go":
        aParent.lastChild.hidden = (count == 0);
        end = count > MAX_HISTORY_MENU_ITEMS ? count - MAX_HISTORY_MENU_ITEMS : 0;

        for (j = count - 1; j >= end; j--) {
          entry = sessionHistory.getEntryAtIndex(j, false);
          if (entry)
            createRadioMenuItem(aParent, j, entry.title, j==index);
        }
      break;

      case "tab":
        aParent.lastChild.hidden = (count > 0);
        end = count > MAX_HISTORY_MENU_ITEMS ? count - MAX_HISTORY_MENU_ITEMS : 0;

        for (j = count - 1; j >= end; j--) {
          entry = sessionHistory.getEntryAtIndex(j, false);
          if (entry)
            createRadioMenuItem(aParent, j, entry.title, j==index);
        }
      break;

      case "global":
        var historyTree = document.getElementById("globalHistoryTree");
        var historyIndex = count = historyTree.treeBoxObject.view.rowCount;
        aParent.childNodes[10].hidden = (count == 0);

        if (count > MAX_HISTORY_MENU_ITEMS)
          count = MAX_HISTORY_MENU_ITEMS;
        if (count == 0)
         return;

        var builder = historyTree.builder.QueryInterface(Components.interfaces.nsIXULTreeBuilder);
        var nameResource = RDF.GetResource("http://home.netscape.com/NC-rdf#Name");

        for (j = count-1; j >= 0; j--) {
          var res = builder.getResourceAtIndex(j);
          var url = res.Value;
          var titleRes = historyTree.database.GetTarget(res, nameResource, true);

          if (titleRes) {
            var title = titleRes.QueryInterface(Components.interfaces.nsIRDFLiteral);
            createRadioMenuItem(aParent, (historyIndex-j), title.Value, (j+historyIndex)==index);
          }
        }
      break;
    }
}

function updateGoMenuPopup(aPopupMenu)
{
  var numberOfTabs = gBrowser.mTabs.length;
  var disabled = (numberOfTabs == 1);
  var menuItems = aPopupMenu.getElementsByAttribute("tbattr", "tabbrowser-multiple");

  for (var i = 0; i < menuItems.length; i++)
    menuItems[i].setAttribute("disabled", disabled);
}

function closeTabOrWindow(aTab)
{
  var tab = (aTab == undefined) ? gBrowser.mCurrentTab : aTab;
  var tabBrowser = getBrowser();
  var tabs = tabBrowser.mTabs.length;
  // 0 = load "about:blank" / 1 = load "about:blank" and hide Tab Bar / 2 = close window & quit browser
  var lastTabBehavior = mzPreferences.readMyPref("multizilla.tabs.last-tab.behavior", "int", 0);
  var aboutBlank = (tab.linkedBrowser.currentURI.spec == "about:blank");

  if (tabs > 1)
    tabBrowser.removeTab(tab);
  else {
    // This is the last tab, so what do we need to do?
    switch(lastTabBehavior) {
      case 0:
      case 1:
        tabBrowser.removeTab(tab);
        // Do we need to hide the Tab Bar?
        if (lastTabBehavior == 1 && tabBrowser.getStripVisibility())
          toggleTabBarVisibility();
        break;
      case 2:
        if (lastTabBehavior == 2 && confirmCloseWS(1))
          BrowserCloseWindow();
    }
  }
}

function preventDupeTab(aURL, useFirstPage, aBrowser, skipReload)
{
  if (mzPreferences.readMyPref("multizilla.prevent.double", "bool", true) && aURL.substring(0,10) != "javascript") {
    var tabHref = null;
    var baseHref = aURL.replace(/#.*/, "");
    var tabBrowser = (aBrowser == undefined) ? getBrowser() : aBrowser;

    if (baseHref[baseHref.length-1] == '/')
      baseHref = baseHref.substring(0, baseHref.length-1);

    for (var i = 0; i < tabBrowser.mTabs.length; i++) {
      if (tabBrowser.mTabs[i].hasAttribute("busy") || useFirstPage) {
        tabHref = tabBrowser.mTabs[i].getAttribute("firstPage");
      }
      else { 
        /*** 
          * 'tabHref' is 'about:blank' when we're restoring a previously saved browser 
          * session, because we need to open new tabs first, before we can modify the 
          * DocShell properties for the corresponding browser!
          */
        tabHref = tabBrowser.mTabs[i].linkedBrowser.webNavigation.currentURI.spec;
      }
      tabHref = tabHref.replace(/#.*/, "");

      if (tabHref[tabHref.length-1] == '/')
        tabHref = tabHref.substring(0, tabHref.length-1);
      var tab = tabBrowser.mTabs[i];

      if (tab && baseHref == tabHref) {
        if (tabBrowser.selectedTab != tab)
          tabBrowser.selectedTab = tab;

        var browser = tab.linkedBrowser;

        if (aURL.indexOf("#") >= 0) {
          /***
            * Anchors like: http://www.foobar.org#123 -> http://www.foobar.org#456 must 
            * use method loadURI() to point the page to the new anchor (a number in this case)
            */
          if (aURL != browser.webNavigation.currentURI.spec) {
            browser.loadURI(aURL);
            browser.reloadWithFlags(Components.interfaces.nsIWebNavigation.LOAD_FLAGS_NONE);
          }
          else 
            browser.reloadWithFlags(Components.interfaces.nsIWebNavigation.LOAD_FLAGS_IS_REFRESH);
        }
        else if (!skipReload) {
          if (!tab.hasAttribute("busy"))
            browser.reloadWithFlags(Components.interfaces.nsIWebNavigation.LOAD_FLAGS_NONE);
        }
        return tab;
      }
    }
  }
  return null;
}

function checkForBlankTab(aBrowser)
{
  if (mzPreferences.readMyPref("multizilla.tabs.blank-tab-first", "bool", true)) {
    var page = "about:blank";
    aBrowser = aBrowser ? aBrowser : getBrowser();

    for (var i = 0; i < aBrowser.mTabs.length; i++) {
      if (aBrowser.mTabs[i].hasAttribute("busy"))
        page = aBrowser.mTabs[i].getAttribute("firstpage");
      else 
        page = aBrowser.mTabs[i].linkedBrowser.currentURI.spec;

      if (page == "about:blank") {
        return aBrowser.mTabs[i].linkedBrowser;
      }
    }
  }
  return null;
}

function openNewTabOrNewWindow()
{
  if (mzPreferences.readMyPref("multizilla.newtabfor.ctrl-n", "bool", true)) {
    var loadInBackground = mzPreferences.readMyPref("browser.tabs.loadInBackground", "bool", true);
    gBrowser.addPreferredTab(!loadInBackground);
    setTimeout("gURLBar.focus();", 0);
  }
  else OpenBrowserWindow();
}

function initMoveTabMenu(aID) 
{
  var windowManagerDS = Components.classes['@mozilla.org/rdf/datasource;1?name=window-mediator'].getService(nsIWindowDataSource);
  var item = document.getElementById(aID);

  while ((item = item.nextSibling)) {
    var url = item.getAttribute('id');
    var win = windowManagerDS.getWindowForResource(url);

    if (win && win == window) {
      item.setAttribute("checked", "true");
      item.setAttribute("disabled", "true");
    }
    else if (!isNavigatorWindow())
      item.setAttribute("disabled", "true");
  }
}

function moveTabToWindow(aTab, aWindowID, aRemoveTabFlag, aInsertFlag)
{
  var windowManagerDS = Components.classes['@mozilla.org/rdf/datasource;1?name=window-mediator'].getService(nsIWindowDataSource);
  var win = windowManagerDS.getWindowForResource(aWindowID);

  if (win && isNavigatorWindow()) {
    var aURL = gBrowser.getBrowserForTab(aTab).currentURI.spec;

    if (aURL) {
      var tabBrowser = win.getBrowser();
      // Add a new blank tab and use it as our targetTab
      var targetTab = tabBrowser.addOrInsertTab(null, null, null, true, aInsertFlag, "duplicateTab");
      // 'tabBrowser.setReloadInterval()' fails without the next line!
      // tabBrowser.selectedTab = targetTab;
      // Get firstPage attribute
      var firstPage = aTab.getAttribute("firstPage");
      /***
        * function relaunchBrowserWindows in quickPrefsMenu.js in one of the 
        * callers of this function, and it checks the 'firstPage' attribute 
        * to remove blank/unused tabs, so we have to verify this attribute!
        */
      if (firstPage == "about:blank")
        firstPage = aURL;
      // Match first page attribute, used for menu item "Tab Home" (Shift+Home)
      targetTab.setAttribute("firstPage", firstPage);
      // Check for a custom label on the source tab
      if (aTab.hasAttribute("customLabel"))
        tabBrowser.setCustomLabel(targetTab, aTab.getAttribute("label")); // Match that label
      // Match reload interval
      tabBrowser.setReloadInterval(targetTab, aTab.mReloadInterval);
      // Match docShell properties
      var sourceBrowser = aTab.linkedBrowser;
      var targetBrowser = targetTab.linkedBrowser;
      tabBrowser.copyDocShellProperties(sourceBrowser, targetBrowser);
      // Match session history
      var tabHistory = gBrowser.getTabHistory(aTab);
      tabBrowser.setTabHistory(targetTab, tabHistory);
      targetBrowser.reloadWithFlags(nsIWebNavigation.LOAD_FLAGS_IS_REFRESH);
      // Do we have to remove the source tab?
      if (aRemoveTabFlag)
        gBrowser.removeTab(aTab);
    }
  }
}

function moveTabToNewWindow(aTab)
{
  var browser = aTab.linkedBrowser;

  if (!browser)
    return;

  var aURL = browser.currentURI.spec;

  if (!aURL)
    return;
  else if (aURL == "about:blank" && browser.sessionHistory.count < 1) {
    var promptService = Components.classes["@mozilla.org/embedcomp/prompt-service;1"].getService(Components.interfaces.nsIPromptService);
    promptService.alert(window, mzGetTextFromBundle("mzWizard"), mzGetTextFromBundle("moveBlankTab"));
    return;
  }
  else {
    var tabBrowser = openWindowAndReturnBrowser();
    // Select targetTab
    var targetTab = tabBrowser.mTabs[0];
    // Match first page attribute, used for menu item "Tab Home" (Shift+Home)
    targetTab.setAttribute("firstPage", aTab.getAttribute("firstPage"));
    // Check for a custom label on the source tab
    if (aTab.hasAttribute("customLabel"))
      tabBrowser.setCustomLabel(targetTab, aTab.getAttribute("label")); // Match that label
    // Match reload interval
    tabBrowser.setReloadInterval(targetTab, aTab.mReloadInterval);
    // Match docShell properties
    tabBrowser.copyDocShellProperties(browser, tabBrowser);
    // Match session history
    var tabHistory = gBrowser.getTabHistory(aTab);
    tabBrowser.setTabHistory(targetTab, tabHistory);
    var targetBrowser = targetTab.linkedBrowser;
    // targetBrowser.reloadWithFlags(nsIWebNavigation.LOAD_FLAGS_IS_REFRESH);
    // It's a move so remove the source tab
    gBrowser.removeTab(aTab);
  }
}

function doFKey(aFunctionKey)
{
  var focusedElement = document.commandDispatcher.focusedElement;

  if (focusedElement instanceof HTMLTextAreaElement || focusedElement instanceof HTMLInputElement 
      /* || focusedElement instanceof HTMLSelectElement */) {
    return;
  }
  aFunctionKey--;

  if (aFunctionKey > -1 && aFunctionKey < gBrowser.mTabs.length)
    gBrowser.selectedTab = gBrowser.mTabs[aFunctionKey];
}

function addNewTabFromClipboard()
{
  var aURL = getLocationFromClipboard();
  if (aURL)
    gBrowser.selectedTab = gBrowser.addTab(aURL);
}

function doReload(aEvent, reloadAll)
{
  aEvent.stopPropagation();

  if (aEvent.ctrlKey || reloadAll)
    gBrowser.reloadAllTabs(aEvent.shiftKey ? true : false);
  else if (aEvent.shiftKey)
    BrowserReloadSkipCache();
  else
    BrowserReload();
}

function doStop(aEvent, stopAll)
{
  aEvent.stopPropagation();

  if (aEvent.ctrlKey || stopAll)
    gBrowser.stopAllTabs();
  else 
    BrowserStop();
}

function initTabbarButtonTooltip(aLabelNode, aCondition, aConditionTrueText, aConditionFalseText)
{
  aLabelNode.value = (aCondition) ? toolbarButton.showTabbarTooltiptext : toolbarButton.hideTabbarTooltiptext;
}

function toggleTabBarVisibility()
{
  dump("\ntoggleTabBarVisibility()");
  var tabBarDisplayed = gBrowser.getStripVisibility();
  dump("\ntoggleTabBarVisibility - tabBarDisplayed: " + tabBarDisplayed);
  var classname = tabBarDisplayed ? "mz-toolbar-hide-button" : "mz-toolbar-show-button";
  document.getElementById("tabstrip-tbbutton").className = classname;
  showHideTabbar();
  var switchTabBarMenuItem = document.getElementById("switchTabBarMenuItem");
  tabBarDisplayed = gBrowser.getStripVisibility();

  if (tabBarDisplayed)
    switchTabBarMenuItem.removeAttribute("disabled");
  else 
    switchTabBarMenuItem.setAttribute("disabled", "true");

  // mzPreferences.writePref("bool", "browser.tabs.autoHide", !tabBarDisplayed);
}

function mzGetNumberOfBlankTabs(aTabBrowser)
{
  var blankTabs = 0;

  for (var i = 0; i < aTabBrowser.mTabs.length; i++) {
    var tab = aTabBrowser.mTabs[i];
    var browser = aTabBrowser.mTabs[i].linkedBrowser;

    if (tab.hasAttribute("busy")) {
      if (tab.getAttribute("firstPage") == "about:blank")
        blankTabs++;
    }
    else if (browser.currentURI.spec == "about:blank")
      blankTabs++;
  }
  return blankTabs;
}

function areAllTabsBlank(aTabBrowser)
{
  if (!aTabBrowser)
    aTabBrowser = getBrowser();

  var numberOfTabs = aTabBrowser.mTabs.length;
  var numberOfBlankTabs = mzGetNumberOfBlankTabs(aTabBrowser);
  return (numberOfTabs == numberOfBlankTabs);
}

function handleNewTabLoadError(aBrowser, aLocation)
{
  // setTimeout(handleTabLoadError, 500, aBrowser, aLocation);
  handleTabLoadError(aBrowser, aLocation);
}

function handleTabLoadError(aBrowser, aLocation)
{
  /* aURL = aLocation.replace(/\/$/, ""); // trim trailing /
  var firstPage = aBrowser.pendingTab.getAttribute("firstPage").replace(/\/$/, ""); // trim trailing /
  // dump("\nfirstPage: " + firstPage + "\naURL: " + aURL);

  if (firstPage != aURL) { // We need this check for 0x80450002 (load aborted) errors!
    aBrowser.pendingTab = null;
  }
  else { */
    var failureAction = 0;
    const failurePrefString = "multizilla.tabs.load-failure.behavior";

    failureAction = mzPreferences.readMyPref(failurePrefString, "int", 2);

    if (failureAction == 1) {
      gBrowser.removeTab(aBrowser.pendingTab);
      mzHideInBrowserHistory(aLocation);
      dump("\nWarning: tab closed, automatically, because it failed to load!");
    }
    else if (failureAction == 2)
    {
      var aEvent = window.document.createEvent('Events');
      aEvent.initEvent("DOMWillOpenModalDialog", false, false);
      aBrowser.contentDocument.defaultView.dispatchEvent(aEvent);

      var dialogTitle = mzGetTextFromBundle("tabFailureDialogTitle");
      var dialogQuestion = mzGetTextFromBundle("tabFailureDialogText");
      var dialogCheckbox = {value:false};
      var dialogCheckboxText = mzGetTextFromBundle("tabFailureDialogCheckboxText");
      var dialogRetryButton = mzGetTextFromBundle("RetryButton");
      var dialogCloseButton = mzGetTextFromBundle("CloseButton");
      var dialogLeaveOpenButton = mzGetTextFromBundle("LeaveOpenButton");
      var promptService = Components.classes["@mozilla.org/embedcomp/prompt-service;1"].getService(Components.interfaces.nsIPromptService);
      var result = promptService.confirmEx(window, dialogTitle, dialogQuestion, promptService.BUTTON_POS_0_DEFAULT +
                                           (promptService.BUTTON_POS_0 * promptService.BUTTON_TITLE_IS_STRING) + 
                                           (promptService.BUTTON_POS_1 * promptService.BUTTON_TITLE_IS_STRING) +
                                           (promptService.BUTTON_POS_2 * promptService.BUTTON_TITLE_IS_STRING),
                                           dialogRetryButton, dialogLeaveOpenButton, dialogCloseButton, dialogCheckboxText, dialogCheckbox);

      aEvent.initEvent("DOMModalDialogClosed", false, false);
      aBrowser.contentDocument.defaultView.dispatchEvent(aEvent);

      if (result == 0) { // Retry button
        aBrowser.pendingTab = null;
        aBrowser.loadURI(aLocation); // aURL);
      }
      else if (result == 1) // Leave Open button selected (or ESC/X)
        aBrowser.pendingTab = null;
      else { // 2 = Close button selected
        gBrowser.removeTab(aBrowser.pendingTab);
        mzHideInBrowserHistory(aLocation);
        dump("\nWarning: tab closed, after prompt, because it failed to load!");
      }
      if (dialogCheckbox.value)
        mzPreferences.writePref("int", failurePrefString, (!result));
    } // else if (failureAction == 2) {
  // } // else {
}

function mzHideInBrowserHistory(aURI)
{
  return; 
  if (typeof(aURI) == "string") {
    const gIOService = Components.classes["@mozilla.org/network/io-service;1"].getService(Components.interfaces.nsIIOService);
    aURI = gIOService.newURI(aURI, null, null);
  }
  var globalHistory;
  // Builds 20040210 and up (see also mozilla bug 224829) 
  if ("@mozilla.org/browser/global-history;2" in Components.classes)
    globalHistory = Components.classes["@mozilla.org/browser/global-history;2"];
  else 
    globalHistory = Components.classes["@mozilla.org/browser/global-history;1"];

  globalHistory.getService(Components.interfaces.nsIBrowserHistory).hidePage(aURI);
}

function mzOpenSearchInTab()
{
  if (mzPreferences.readMyPref("multizilla.newtabfor.websearch", "bool", true)) {
    return gBrowser.addTab(null, null, null, true, null, null, null, true);
  }
  return null;
}

function mzOpenSearch(aSearchStr)
{
  /* See also: http://lxr.mozilla.org/mozilla1.7/source/xpfe/browser/resources/content/navigator.js#999
     and:      http://lxr.mozilla.org/mozilla/source/xpfe/browser/resources/content/navigator.js#1034 */
  var newWindowFlag = (mzOpenSearchInTab() == null);
  var powerSearch;

  try {
    pref.getIntPref("browser.search.powermode");
    powerSearch = true;
  } catch(ex) {
    powerSearch = false;
  }
  if (powerSearch)
    OpenSearch('internet', false, aSearchStr, newWindowFlag);
  else 
    OpenSearch('internet', aSearchStr, newWindowFlag);
}

function mzOpenTabOrWindow(aEvent)
{
  if (aEvent.button == 1 || aEvent.ctrlKey) {
    if (aEvent.shiftKey)
      window.open();
    else
      OpenBrowserWindow();
  }
  else if (aEvent.button == 0) {
    var loadInBackground = mzPreferences.readMyPref("browser.tabs.loadInBackground", "bool", true);

    if (aEvent.shiftKey)
      loadInBackground = !loadInBackground;

    gBrowser.addPreferredTab(!loadInBackground);
    setTimeout('gURLBar.focus();', 0);
  }
}

